import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="backtest_pkg",
    version="0.1.3",
    author="Andy Hu",
    author_email="andyhu2014@gmail.com",
    description="A package for backtesting portfolio strategies and trading signals",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/andyhu4023/backtest_pkg.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)